# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/05c_data.gym.ipynb (unless otherwise specified).

__all__ = ['source_events', 'return_data', 'Source', 'ReturnHistory', 'TstCallback', 'np_type_cast', 'GymLoop',
           'FirstLastTfm', 'FirstLast', 'Reward', 'NEpisodes']

# Cell
# Python native modules
import os
from collections import deque
from copy import deepcopy
from time import sleep
from typing import *
# Third party libs
from fastcore.all import *
from fastai.torch_basics import *
from fastai.data.all import *
from fastai.basics import *
from fastai.callback.all import *
from torch.utils.data import Dataset
from torch import nn
import torch
import gym
try:
    import pybulletgym
except ModuleNotFoundError as e:pass
import numpy as np
# Local modules
from ..core import *
from ..callback.core import *
from .block_simple import *
from ..agent import *

# Cell
_loop=L(['event.after_create','Start Setup','event.initialize','End Setup',
             'event.before_episodes',
             'Start Episodes',
                 'event.reset',
                 'event.do_action',
                 'event.do_step',
                 'event.do_render',
                 'event.before_history',
                 'event.history',
                 'event.after_history',
             'End Episodes',
             'event.after_episodes'
             ])

mk_class('source_events', **parse_events(_loop).map_dict(),
         doc="All possible events as attributes to get tab-completion and typo-proofing")

#nbdev_comment _all_=['source_events']

# Cell
def return_data(o:dict):
    source,data=o['source'],o['history']

    return {'source':source,'history':data}

class Source(Loop):
    _loop=_loop
    _events=source_events
    _default='source'
    end_event=parse_events(_loop)[-1]

    @delegates(Loop)
    def __init__(self,cbs=None,test_mode=False,**kwargs):
        self.idx=0
        self.data_fields='state,next_state,done,all_done,env_id,worker_id'\
                           ',action,episode_id,accum_rewards,reward,step'.split(',')
        self.ignore_fields=[]
        self.test_field=torch.full((1,5),self.idx)
        self.return_fn=return_data
        self.loop_history_yield=False
        store_attr(but='cbs',state=None,next_state=None,done=None,all_done=None,
                   env_id=0,action=None,episode_id=0,accum_rewards=0,reward=0,step=0,
                   skip_history_return=False)
        super().__init__(cbs=cbs,**kwargs)

    def after_create(self):
        self('initialize')
        return self

    def _history(self):
        self.loop_history_yield=False
        self('history')
        if self.test_mode:
            self.this=torch.full((1,5),self.idx)
            if 'test_field' not in self.data_fields: self.data_fields.append('test_field')
        return self.return_fn(dict(source=self,history=self.data()))['history']

    def data(self)->BD:
        return BD({s:(ifnone(getattr(self,s,None),TensorBatch([[0]])) if self.test_mode else getattr(self,s))
                for s in self.data_fields if not in_(s,self.ignore_fields)})

    def __iter__(self):
        self('before_episodes')
        while True:
            self.idx+=1
            self('reset')
            self('do_action')
            self('do_step')
            if self.test_mode: self.test_field=torch.full((1,5),self.idx)
            self('do_render')
            self('before_history')
            if not self.skip_history_return: yield self._history()
            while self.loop_history_yield:   yield self._history()
            self('after_history')

# Cell
class ReturnHistory(Transform):
    def encodes(self,o:dict):
        source,data=o['source'],o['history']
        history=sum(source.histories)
        if len(source.histories)==1 and sum(history['done'])>0:
            source.all_done=TensorBatch([[True]])
            history['all_done']=TensorBatch([[True]])
        source.histories.popleft()
        return {'source':source,'history':history}

# Cell
class TstCallback(AgentCallback):
    def __init__(self,action_space=None,constant=None): store_attr()
    def before_noise(self):
        bs=self.experience['state'].shape[0]
        self.agent.action=Tensor([[self.constant] if self.constant is not None else
                                   self.action_space.sample()
                                   for _ in range(bs)])
        self.agent.experience=D(merge(self.experience,{'random_action':np.random.randint(0,3,(bs,1))}))

def np_type_cast(a,dtype): return a.astype(dtype)

class GymLoop(LoopCallback):
    _methods=source_events
    _default='source'
    def __init__(self,env_name:str,agent=None,mode=None,
                 steps_count:int=1,steps_delta:int=1,seed:int=None,
                 worker_id:int=0,dtype=None):
        store_attr()

    def initialize(self):
        self.source.histories=deque(maxlen=self.steps_count)
        self.source.return_fn=Pipeline([ReturnHistory])
        if self.mode!='rgb_array': self.source.ignore_fields.append('image')
        self.source.worker_id=get_worker_info()
        self.source.worker_id=TensorBatch([[getattr(self.worker_id,'id',0)]])

        self.source.env=gym.make(self.env_name)
        if isinstance(self.env.action_space,gym.spaces.Discrete):
            self.dtype=int
        elif isinstance(self.env.action_space,gym.spaces.Box):
            if issubclass(self.env.action_space.dtype.type,(float,np.floating)):
                self.dtype=partial(np_type_cast,dtype=float)
            if issubclass(self.env.action_space.dtype.type,(int,np.integer)):
                self.dtype=partial(np_type_cast,dtype=int)
        else:
            self.dtype=float

        self.agent=ifnone(self.agent,Agent(cbs=TstCallback(action_space=self.env.action_space)))
        self.source.episode_id=TensorBatch([[0]])
        self.source.done=TensorBatch([[True]])
        self.source.all_done=TensorBatch([[True]])

    def reset(self):
        if len(self.histories)==0 and sum(self.all_done)>=1:
            self.source.env_id=TensorBatch([[0]])
            self.source.reward=TensorBatch([[0.0]])
            self.source.episode_id+=1
            self.source.accum_rewards=TensorBatch([[0.0]])
            self.source.step=TensorBatch([[0]])
            self.source.env_id=TensorBatch([[0]])
            self.source.done=TensorBatch([[False]])
            self.source.all_done=TensorBatch([[False]])
            self.source.env.seed(self.seed)
            self.source.state=TensorBatch(self.env.reset()).unsqueeze(0)

    def before_episodes(self): self('initialize')

    def do_step(self):
        action,experience=self.agent.do_action(**self.data())
        for k,v in experience.items():
            if v is not None: setattr(self.source,k,TensorBatch(v))
        self.source.action=TensorBatch(action)
        step_action=self.action.numpy()[0]
        next_state,reward,done,_=self.env.step(self.dtype(step_action))
        self.source.next_state=TensorBatch(next_state).unsqueeze(0)
        self.source.reward=TensorBatch([[reward]])
        self.source.accum_rewards+=reward
        self.source.done=TensorBatch([[done]])
        self.source.step+=1

    def before_history(self):
        self.source.skip_history_return=False
        self.histories.append(deepcopy(self.data()))

        if self.done.sum().item()<1:
            self.source.skip_history_return=len(self.histories)<self.steps_count or \
                                            int(self.step.item())%self.steps_delta!=0

    def history(self):
        if sum(self.done)>0 and len(self.histories)>1 and self.steps_count>1:
            self.source.loop_history_yield=True

    def after_history(self):
        self.source.state=self.source.next_state


add_docs(GymLoop,"Handles iterating through single openai gym (and varients).",
         before_history="""Primarily appends `self.data()` to `self.histories` however it also...

         If the number of elements in `self.histories` is empty, or about to be empty, and
         `self.done==True` it sets the `self.all_done` field.

         If `self.done==True` also determine if we should skip returning a history.
         Mainly if we need to accumulate the histories first.
         """,
         initialize="Sets up most of the needed fields as well as the environment itself.",
         reset="Resets the env and fields if the histories have been emptied.",
         before_episodes="Call the initialization method again.",
         do_step="Get actions from the agent and take a step through the env.",
         history="If the environment is done, we want to loop through the histories and empty them.",
         after_history="Once everything is processed, we set the next state to the current state for the next loop.")

# Cell
class FirstLastTfm(Transform):
    def __init__(self,gamma):
        store_attr()
        super().__init__()


    def encodes(self,o):
#         print('encoding')
        source,history=o['source'],o['history']
#         print(source,history,o)
        element=history[0]
        if history.bs()!=1:
            remainder=history[1:]
            reward=element['reward']
            for e in reversed(remainder['reward']):
                reward*=self.gamma
                reward+=e
            element['reward']=reward
            element['next_state']=history[-1]['next_state']
            element['done']=history[-1]['done']

#         print(source,element)
        return {'source':source,'history':element}

class FirstLast(LoopCallback):
    _methods=source_events
    _default='source'

    def __init__(self,gamma=0.99): store_attr()
    def before_episodes(self): self('initialize')
    def initialize(self):
        if isinstance(self.source.return_fn,Pipeline):
            self.source.return_fn.add(FirstLastTfm(self.gamma))
        else:
            self.source.return_fn=Pipeline([self.source.return_fn,FirstLastTfm(self.gamma)])

# Cell
class Reward(Metric):
    order=30
    reward=None
    rolling_reward_n=100
    keep_rewards=False

    def reset(self):
        if self.reward is None: self.reward=deque(maxlen=self.rolling_reward_n)

    def accumulate(self,learn):
        xb=learn.xb[0]
        if xb['all_done'].sum()>0:
            final_rewards=to_detach(xb['accum_rewards'][xb['all_done']])
            for i in final_rewards.reshape(-1,).numpy().tolist(): self.reward.append(i)

    @property
    def value(self): return np.average(self.reward) if len(self.reward)>0 else 0

# Cell
class NEpisodes(Metric):
    order=30
    n_episodes=None

    def reset(self):
        if self.n_episodes is None: self.n_episodes=0

    def accumulate(self,learn):
        xb=learn.xb[0]
        if xb['all_done'].sum()>0:
            new_episodes=torch.unique(to_detach(xb['episode_id'][xb['all_done']]))
            self.n_episodes+=len(new_episodes)


    @property
    def value(self): return self.n_episodes


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/05b_data.block_simple.ipynb (unless otherwise specified).

__all__ = ['parse_events', 'source_events', 'Source', 'SourceDataset', 'VerboseChecked', 'IterableTfmdLists',
           'IterableDatasets', 'IterableDataBlock', 'TransformBlock', 'Source', 'vstack_collate', 'fr_collate',
           'IterableTfmdDL', 'get_sources', 'SourceDataBlock']

# Cell
# Python native modules
import os
from collections import deque
from copy import deepcopy
from time import sleep
# Third party libs
from fastcore.all import *
from fastai.torch_basics import *
from fastai.data.all import *
from fastai.basics import *
from fastai.callback.all import *
from torch.utils.data import Dataset
from torch import nn
import torch
import gym
import numpy as np
# Local modules
from ..core import *
from ..callback.core import *
from ..agent import *

# Cell
def parse_events(loop:L):
    return loop.filter(lambda s:in_('event.',s)).map(lambda s:s.replace('event.',''))

# Cell
_loop=L(['event.after_create','Start Setup','event.initialize','End Setup',
             'event.before_episodes',
             'Start Episodes',
                 'event.reset',
                 'event.do_action',
                 'event.do_step',
                 'event.render',
                 'event.history',
             'End Episodes',
             'event.after_episodes'
             ])

mk_class('source_events', **parse_events(_loop).map_dict(),
         doc="All possible events as attributes to get tab-completion and typo-proofing")

#nbdev_comment _all_=['source_events']

# Cell
class Source(Loop):
    _loop=_loop
    _events=source_events
    _default='source'
    end_event=parse_events(_loop)[-1]

    @delegates(Loop)
    def __init__(self,cbs=None,**kwargs):
        super().__init__(cbs=cbs,**kwargs)
        store_attr(but='cbs')
        self.idx=0

    def after_create(self):
        self('initialize')
        return self

    def __iter__(self):
        while True:
            self.idx+=1
            self('do_step')
            yield {'this':torch.full((1,5),self.idx)}

# Cell
class SourceDataset(IterableDataset):
    "Iterates through a `source` object. Allows for re-initing source connections when `num_workers>0`"
    def __init__(self,source=None): self.source=source
    def __iter__(self):
        source=iter(self.source)
        yield from source

# Cell
class VerboseChecked(LoopCallback):
    _methods=source_events
    def __init__(self,show_worker_id=True,show_env_id=True):store_attr()

    def initialize(self):
        worker_id=get_worker_info()
        worker_id=worker_id.id if worker_id is not None else 0
        if self.show_worker_id: print('Worker id: ',worker_id)

        self.env=gym.make('HumanoidPyBulletEnv-v0')
        self.env.reset()
    def do_step(self):
        if self.show_env_id: print('Env Id: ',id(self.env))

# Cell
from itertools import cycle

class IterableTfmdLists(TfmdLists):
    def __iter__(self): return cycle(self[i] for i in range(len(self)))

class IterableDatasets(Datasets):
    def __iter__(self): return cycle(self[i] for i in range(len(self)))

# Cell
class IterableDataBlock(DataBlock):
    tl_type = TfmdLists
    datasets_type = Datasets

    @delegates(DataBlock)
    def __init__(self,blocks=None,datasets_type=None,n_inp=None,**kwargs):
        blocks = L(self.blocks if blocks is None else blocks)
        for b in blocks:
            if getattr(b, 'tl_type', None) is not None: self.tl_type = b.tl_type
        if datasets_type is not None: self.datasets_type=datasets_type
        if (not is_listy(blocks) or len(blocks)==1) and n_inp is not None: n_inp=0
        super().__init__(blocks=blocks,n_inp=n_inp, **kwargs)

    def datasets(self, source, verbose=False):
        self.source = source                     ; pv(f"Collecting items from {source}", verbose)
        items = (self.get_items or noop)(source) ; pv(f"Found {len(items)} items", verbose)
        splits = (self.splitter or RandomSplitter())(items)
#         pv(f"{len(splits)} datasets of sizes {','.join([str(len(s)) for s in splits])}", verbose)
        return self.datasets_type(items, tfms=self._combine_type_tfms(), splits=splits, dl_type=self.dl_type,
                                  n_inp=self.n_inp, verbose=verbose, tl_type=self.tl_type)

@patch
def __init__(self:Datasets, items=None, tfms=None, tls=None, n_inp=None, dl_type=None,tl_type=TfmdLists, **kwargs):
    super(Datasets,self).__init__(dl_type=dl_type)
    self.tls = L(tls if tls else [tl_type(items, t, **kwargs) for t in L(ifnone(tfms,[None]))])
    self.n_inp = ifnone(n_inp, max(1, len(self.tls)-1))

class TransformBlock():
    "A basic wrapper that links defaults transforms for the data block API"
    def __init__(self, type_tfms=None, item_tfms=None, batch_tfms=None, dl_type=None, dls_kwargs=None, tl_type=None):
        self.type_tfms  =            L(type_tfms)
        self.item_tfms  = ToTensor + L(item_tfms)
        self.batch_tfms =            L(batch_tfms)
        self.dl_type,self.dls_kwargs = dl_type,({} if dls_kwargs is None else dls_kwargs)
        self.tl_type = tl_type

# Cell
class Source(Loop):
    _loop=_loop
    _events=source_events
    _default='source'
    end_event=parse_events(_loop)[-1]

    @delegates(Loop)
    def __init__(self,cbs=None,**kwargs):
        super().__init__(cbs=cbs,**kwargs)
        store_attr(but='cbs')
        self.idx=0

    def after_create(self):
        self('initialize')
        print('init')
        return self

    def __iter__(self):
        while True:
            self.idx+=1
            self('do_step')
            yield {'this':torch.full((1,5),self.idx)}

# Cell
from torch.utils.data._utils.collate import default_collate_err_msg_format
from torch.utils.data._utils.collate import *

def vstack_collate(batch):
    "99% similar to default_collate, however vstacks tensors thus assuming they already have a batch dim"

    elem = batch[0]
    elem_type = type(elem)
    if isinstance(elem, torch.Tensor):
        out = None
        if torch.utils.data.get_worker_info() is not None:
            # If we're in a background process, concatenate directly into a
            # shared memory tensor to avoid an extra copy
            numel = sum([x.numel() for x in batch])
            storage = elem.storage()._new_shared(numel)
            out = elem.new(storage)
        return torch.vstack(batch, out=out)
    elif elem_type.__module__ == 'numpy' and elem_type.__name__ != 'str_' \
            and elem_type.__name__ != 'string_':
        if elem_type.__name__ == 'ndarray' or elem_type.__name__ == 'memmap':
            # array of string classes and object
            if np_str_obj_array_pattern.search(elem.dtype.str) is not None:
                raise TypeError(default_collate_err_msg_format.format(elem.dtype))

            return vstack_collate([torch.as_tensor(b) for b in batch])
        elif elem.shape == ():  # scalars
            return torch.as_tensor(batch)
    elif isinstance(elem, float):
        return torch.tensor(batch, dtype=torch.float64)
    elif isinstance(elem, int_classes):
        return torch.tensor(batch)
    elif isinstance(elem, string_classes):
        return batch
    elif isinstance(elem, container_abcs.Mapping):
        return {key: vstack_collate([d[key] for d in batch]) for key in elem}
    elif isinstance(elem, tuple) and hasattr(elem, '_fields'):  # namedtuple
        return elem_type(*(default_collate(samples) for samples in zip(*batch)))
    elif isinstance(elem, container_abcs.Sequence):
        # check to make sure that the elements in batch have consistent size
        it = iter(batch)
        elem_size = len(next(it))
        if not all(len(elem) == elem_size for elem in it):
            raise RuntimeError('each element in list of batch should be of equal size')
        transposed = zip(*batch)
        return [vstack_collate(samples) for samples in transposed]

    raise TypeError(default_collate_err_msg_format.format(elem_type))

# Cell
_collate_types = (ndarray, Tensor, typing.Mapping, str)

def fr_collate(t):
    "A replacement for PyTorch `default_collate` which maintains types and handles `Sequence`s"
    b = t[0]
    return (vstack_collate(t) if isinstance(b, _collate_types)
            else type(t[0])([fr_collate(s) for s in zip(*t)]) if isinstance(b, Sequence)
            else vstack_collate(t))

# Cell

from fastai.data.load import _loaders

class IterableTfmdDL(TfmdDL):
    def __init__(self, dataset,bs=64, shuffle=False, num_workers=None, verbose=False, do_setup=True,**kwargs):
        super().__init__(dataset, bs=bs, shuffle=shuffle, num_workers=num_workers,**kwargs)
        self._loader=None

    def create_batch(self, b): return (fr_collate,fa_convert)[self.prebatched](b)

# Cell
def get_sources(_,ls): return [o.after_create() for o in ls]

class SourceDataBlock(IterableDataBlock):
    datasets_type=IterableDatasets
    get_items=get_sources
    blocks=TransformBlock(
        type_tfms=[lambda o:iter(o)],
        item_tfms=lambda o:next(o),
        tl_type=IterableTfmdLists,
        dl_type=IterableTfmdDL,
        dls_kwargs={'indexed':False,'shuffle':False,'persistent_workers':True})

# Cell
@patch
def _do_epoch_validate(self:Learner,*args,**kwargs): return 0

# Cell
@patch
def after_create(self:Callback):
    for cb in self.learn.cbs:
        if hasattr(cb,'train_metrics'): cb.train_metrics=True


# Cell
@patch
def after_batch(self:Recorder):
    "Update all metrics and records lr and smooth loss in training"
    if len(self.yb) == 0 and len(self.xb) == 0: return
    mets = self._train_mets if self.training else self._valid_mets
    for met in mets: met.accumulate(self.learn)
    if not self.training: return
    self.lrs.append(self.opt.hypers[-1]['lr'])
    self.losses.append(self.smooth_loss.value)
    self.learn.smooth_loss = self.smooth_loss.value



# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/10e_agents.dqn.categorical.ipynb (unless otherwise specified).

__all__ = ['create_support', 'CategoricalDQN', 'distribute', 'final_distribute', 'categorical_update',
           'show_q_distribution', 'PartialCrossEntropy', 'CategoricalDQNTrainer', 'CategoricalArgMaxFeed', 'show_q',
           'show_q_and_max_distribution']

# Cell
# Python native modules
from typing import *
# Third party libs
import torch
from torch.nn import *
from fastcore.all import *
from fastai.learner import *
from fastai.torch_basics import *
from fastai.torch_core import *
from fastai.callback.all import *
# Local modules
from ...data.block_simple import *
from ...data.gym import *
from ...agent import *
from ...core import *
from .core import *
from .targets import *

# Cell
def create_support(v_min=-10,v_max=10,n_atoms=51)->Tuple[Tensor,float]:
    "Creates the support and returns the z_delta that was used."
    z_delta=(v_max-v_min)/(n_atoms-1)
    return (Tensor([i*z_delta for i in range(n_atoms)])+v_min,z_delta)

# Cell
class CategoricalDQN(Module):
    def __init__(self,state_sz:int,action_sz:int,n_atoms:int=51,hidden=512,
                 v_min=-10,v_max=10):
        store_attr()
        self.layers=Sequential(
            Linear(state_sz,hidden),
            ReLU(),
            Linear(hidden,action_sz*n_atoms),
        )
        self.supports,self.z_delta=create_support(v_min=v_min,v_max=v_max,n_atoms=n_atoms)
        self.softmax=Softmax(dim=2)

    def to(self, *args, **kwargs):
        self = super().to(*args, **kwargs)
        self.supports=self.supports.to(*args, **kwargs)
        return self

    def forward(self,x):
        return self.layers(x).view(x.shape[0],self.action_sz,self.n_atoms)
    def policy(self,x): return (self.supports*self.p(x)).mean(dim=2)
    def q(self,x): return (self.supports*self.p(x)).sum(dim=2)
    def p(self,x): return self.softmax(self(x))

# Cell
def distribute(projection,left,right,support_value,p_a,atom,done):
    "Does: m_l <- m_l + p_j(x_{t+1},a*)(u - b_j) operation for non-final states."
    diffs=torch.hstack((support_value-left,right-support_value))
    # If they are the same location, then just split the value in half, and add twice
    diffs[(left==right).reshape(-1,),:]=0.5
    mask=~done.reshape(-1,)

    left_v=projection[mask].gather(1,left[mask])+(p_a[:,atom][mask]*diffs[:,0][mask]).reshape(-1,1)
    right_v=projection[mask].gather(1,right[mask])+(p_a[:,atom][mask]*diffs[:,1][mask]).reshape(-1,1)

    projection[mask]=projection[mask].scatter(dim=1,index=left[mask],src=left_v)
    projection[mask]=projection[mask].scatter(dim=1,index=right[mask],src=right_v)

def final_distribute(projection,left,right,support_value,p_a,atom,done):
    "Does: m_l <- m_l + p_j(x_{t+1},a*)(u - b_j) operation for final states."
    diffs=torch.hstack((support_value-left,right-support_value))
    # If they are the location, then just split the value in half, and add twice
    diffs[(left==right).reshape(-1,),:]=0.5
    mask=done.reshape(-1,)

    left_v=diffs[:,0].reshape(-1,1)
    right_v=projection[mask].gather(1,right)+diffs[:,1].reshape(-1,1)

    projection[mask]=0.0
    projection[mask]=projection[mask].scatter(dim=1,index=left,src=left_v)
    projection[mask]=projection[mask].scatter(dim=1,index=right,src=right_v)

# Cell
def categorical_update(support,delta_z,q,p,actions,rewards,dones,v_min=-10,
                       v_max=10,n_atoms=51,gamma=0.99,passes=None):
    bs=q.shape[0]
    passes=ifnone(passes,n_atoms)
    # Do this outside of the loop so we only have to do it once
    # Represents: p_j(x_{t+1},a*)
    p_a=p[torch.arange(bs),actions.reshape(-1,)]#.reshape(-1,1)
    # get a*
    next_actions=torch.argmax(q,dim=1)
    # m_i = 0 for i in [0,N-1]
    projection=torch.zeros((bs,n_atoms)).to(device=default_device())
    # j in [0, N - 1]
    for atom in range(0,passes):
        # Tz_j <- [r_t + gamma * z_j]_v_min^v_max
        target_z=rewards+gamma*support[atom]
        target_z[dones.reshape(-1)]=rewards[dones.reshape(-1)].float()
#         print(rewards,gamma*support[atom])
        target_z=torch.clamp(target_z,v_min,v_max)
        # b_j <- (Tz_j - Vmin)/delta_z
        support_value=(target_z-v_min)/delta_z
        # l <- floor(b_j), u <- ceil(b_j)
        left,right=support_value.floor().long(),support_value.ceil().long()
        # m_l <- m_l + p_j(x_{t+1},a*)(u - b_j)
        distribute(projection,left,right,support_value,p_a,atom,dones)

    if dones.sum()>=1:
        target_z=rewards[dones.reshape(-1)]
        target_z=torch.clamp(target_z,v_min,v_max)
        support_value=(target_z-v_min)/delta_z
        left,right=support_value.floor().long(),support_value.ceil().long()
        final_distribute(projection,left,right,support_value,p_a,atom,dones)
    return projection

# Cell
def show_q_distribution(cat_dist,title='Update Distributions'):
    "`cat_dist` being shape: (bs,n_atoms)"
    from IPython.display import HTML
    import plotly.graph_objects as go
    fig = go.Figure(data=[go.Surface(z=to_detach(cat_dist).numpy())])
    fig.update_layout(title=title,autosize=False,
                      width=500, height=500,#template='plotly_dark',
                      margin=dict(l=65, r=50, b=80, t=90))
    return HTML(fig.to_html())

# Cell
def PartialCrossEntropy(p,q): return (-p*q).sum(dim=1).mean()

# Cell
class CategoricalDQNTrainer(Callback):

    def __init__(self,n_batch=0,target_sync=300,discount=0.99,n_steps=1):
        store_attr()
        self._xb=None
        self.local_xb=None
        self.local_pred=None

    def before_fit(self):
        self.learn.target_model=deepcopy(self.learn.model.model)
        self.n_batch=0

    def after_pred(self):
        self.learn.yb=self.xb
        self.learn.xb=self.xb[0]

        self._xb=({k:v.clone() for k,v in self.xb.items()},)

        with torch.no_grad():
            distribution_m=categorical_update(self.learn.target_model.supports,
                                      self.learn.target_model.z_delta,
                                      self.learn.target_model.q(self.xb['next_state']),
                                      self.learn.target_model.p(self.xb['next_state']),
                                      self.xb['action'],self.xb['reward'],
                                      self.xb['done'])
        v=self.learn.model.model(self.xb['state'])
        self.local_v=v
        self.learn.pred=v[np.arange(v.shape[0]),self.xb['action'].reshape(-1,),:]
        self.learn.pred=F.log_softmax(self.learn.pred,dim=1)
        self.learn.xb=(distribution_m,)

        self.local_xb=self.learn.xb
        self.local_pred=self.learn.pred

    def before_backward(self): self.learn.xb=self._xb

    def after_batch(self):
        if self.n_batch%self.target_sync==0:
            self.target_model.load_state_dict(self.learn.model.state_dict())
        self.n_batch+=1

# Cell
class CategoricalArgMaxFeed(AgentCallback):
    def before_action(self):
        raw_action=self.agent.model.policy(self.experience['state'].to(default_device()))
        self.agent.raw_action_shape=raw_action.shape
        self.agent.action=torch.argmax(raw_action,dim=1).reshape(-1,1).to(default_device())

# Cell
def show_q(cat_dist,title='Update Distributions'):
    "`cat_dist` being shape: (bs,n_atoms)"
    from IPython.display import HTML
    import plotly.graph_objects as go
    from plotly.subplots import make_subplots

    distributions=to_detach(cat_dist).numpy()
    actions=np.argmax(distributions,axis=1).reshape(-1,)
    fig = make_subplots(rows=1, cols=2,
                        specs=[[{"type": "surface"},{"type": "xy"}]])
    fig.add_trace(go.Surface(z=distributions, showscale=False),row=1, col=1)
    fig.add_trace(go.Scatter(x=np.arange(len(actions)),y=actions),row=1, col=2)
    fig.update_layout(title=title,autosize=False,
                      width=1000, height=500,#template='plotly_dark',
                      margin=dict(l=65, r=50, b=80, t=90))
    return HTML(fig.to_html())

# Cell
def show_q_and_max_distribution(cat_dist,title='Update Distributions'):
    "`cat_dist` being shape: (bs,n_atoms)"
    from IPython.display import HTML
    import plotly.graph_objects as go
    from plotly.subplots import make_subplots

    distributions=to_detach(cat_dist).numpy()
    actions=np.argmax(distributions,axis=1).reshape(-1,)
    fig = make_subplots(rows=1, cols=2,
                        specs=[[{"type": "surface"},{"type": "xy"}]])
    fig.add_trace(go.Surface(z=distributions, showscale=False),row=1, col=1)
    fig.add_trace(go.Scatter(x=np.arange(len(actions)),y=actions),row=1, col=2)
    fig.update_layout(title=title,autosize=False,
                      width=1000, height=500,#template='plotly_dark',
                      margin=dict(l=65, r=50, b=80, t=90))
    return HTML(fig.to_html())


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/05a_data.block.ipynb (unless otherwise specified).

__all__ = ['TstCallback', 'source_events', 'Src', 'SrcCallback', 'GymSrc', 'SrcProfiler', 'FirstLast', 'SourceDataset',
           'IterableTfmdLists', 'vstack_collate', 'fr_collate', 'IterableTfmdDL', 'IterableDataBlock', 'TransformBlock',
           'SourceBlock', 'map_iters', 'noop_splitter', 'get_all_keys', 'SourceDataBlock']

# Cell
# Python native modules
import os
from collections import deque
from copy import deepcopy
from time import sleep
# Third party libs
from fastcore.all import *
from fastai.torch_basics import *
from fastai.data.all import *
from fastai.basics import *
from torch.utils.data import Dataset
from torch import nn
import torch
import gym
import numpy as np
# Local modules
from ..core import *
from ..callback.core import *
from ..agent import *

# Cell
class TstCallback(AgentCallback):
    def __init__(self,action_space=None,constant=None): store_attr()
    def before_noise(self):
        bs=self.experience['state'].shape[0]
        self.agent.action=Tensor([[ifnone(self.constant,self.action_space.sample())] for _ in range(bs)])
        self.agent.experience=D(merge(self.experience,{'random_action':np.random.randint(0,3,(bs,1))}))

# Cell
_events=L.split('create reset do_action do_step render history episodes initialize')
_events= _events.map(lambda s:'before_'+s) + _events.map(lambda s:'after_'+s)+_events.map(lambda s:'after_cancel_'+s)
_events+=L.split('reset do_step render initialize history do_action')

mk_class('source_events', **_events.map_dict(),
         doc="All possible events as attributes to get tab-completion and typo-proofing")

#nbdev_comment _all_=['source_events']

_loop=L(['Start Setup','before_initialize','initialize','after_initialize','End Setup',
             'before_episodes',
             'Start Episodes','before_reset','reset','after_reset',
                 'before_do_action','do_action','after_do_action',
                 'before_do_step','do_step','after_do_step',
                 'before_render','render','after_render',
                 'before_history','history','after_history',
             'End Episodes',
             'after_episodes'
             ])
_default_shapes={'done':(1,1),'all_done':(1,1),'step':(1,1),'env_id':(1,1),
                 'p_id':(1,1),'episode_num':(1,1),'accum_rewards':(1,1)}

# Cell
class Src(Loop):
    _loop=_loop
    _events=source_events
    _default='source'
    end_event='after_episodes'
    initial_experience_flds='done step state next_state env_id'

    @delegates(Loop)
    def __init__(self,env:object,agent=None,n_envs:int=1,steps_count:int=1,steps_delta:int=1,
                 seed:int=None,mode=None,num_workers=0,but='',cbs=None,shape_map=None,**kwargs):
        super().__init__(cbs=cbs,**kwargs)
        store_attr(but='cbs,shape_map')
        self.env_kwargs=kwargs
        self.pool=L()
        self.histories=L()
        self.current_history=L()
        self.shape_map=merge(_default_shapes,ifnone(shape_map,{}))
        self.all_exp:BD=None
        self.imask=torch.zeros((1,))
        if self.mode is None: self.but+=',image'
        # Fields
        self.done=None
        self.all_done=None
        self.step=None
        self.env_id=None
        self.p_id=None

    def _init_state(self):     self('initialize')
    def _do_reset(self):       self('reset')
    def _do_action(self):      self('do_action')
    def _do_step(self):        self('do_step')
    def _do_render(self):      self('render')
    def process_history(self,history):
        self.current_history=history
        self('history')
        return self.current_history

    def yield_histories(self):
        self('before_history')
        for i,idx in enumerate(self.imask.nonzero().reshape(-1,)):
            self.histories[idx].append(deepcopy(self.active_exp()[i]))
            if len(self.histories[idx])==self.steps_count and int(self.step[i])%self.steps_delta==0 and not bool(self.done.reshape(-1,)[i]):
                yield self.process_history(sum(self.histories[idx]))
            if bool(self.done.reshape(-1,)[i]):
#                 print(self.histories[idx],'\n')
#                 if 0<len(self.histories[idx])<self.steps_count:
#                     yield self.process_history(sum(self.histories[idx]))
                while len(self.histories[idx])>=1:
                    yield self.process_history(sum(self.histories[idx]))
                    self.histories[idx].popleft()
        self('after_history')

    def active_exp(self,bs=None):
        return BD({k:TensorBatch(getattr(self,k),bs=ifnone(bs,self.imask.sum()))
                   for k in self.shape_map})

    def __iter__(self):
        "Iterates through a list of environments."
        if not self.pool:self._with_events(self._init_state,'initialize',Exception)
        self('before_episodes')
        while True:
            self._with_events(self._do_reset,'reset',Exception)
            self._with_events(self._do_action,'do_action',Exception)
            self._with_events(self._do_step,'do_step',Exception)
            if self.mode is not None: self._with_events(self._do_render,'render',Exception)
            yield from self.yield_histories()
        self('after_episodes')

# Cell
class SrcCallback(LoopCallback):
    _default='source'
    _methods=_events

    def after_cancel_initialize(self):raise
    def after_cancel_step(self):      raise
    def after_cancel_do_action(self): raise
    def after_cancel_do_step(self):   raise
    def after_cancel_episodes(self):  raise
    def after_cancel_episode(self):   raise
    def after_cancel_history(self):   raise
    def after_cancel_render(self):    raise
    def after_cancel_create(self):    raise
    def after_cancel_reset(self):     raise

# Cell
def _add_batch(t):
    if len(t)==1:
        if type(t[0])==float: return (1,1)
        else:                 return (1,t[0])
    elif t[0]!=1: return (1,*t)
    return t

def _fix_shape_map(t:Tensor,expected_shape,bs=None):
    if not hasattr(t,'shape'): t=Tensor([t])
    if bs is not None:
        expected_shape=(bs,*expected_shape[1:])
    if expected_shape!=tuple(t.shape):
        if isinstance(t,np.ndarray): t=Tensor(t)
        t=t.unsqueeze(0)
        test_eq(expected_shape,t.shape)
    return t

def _env_seed(o,seed): return o.seed(seed)

def _env_reset(o,shape_map=None):
    state=o.reset()
    return _fix_shape_map(state,shape_map['state'])

def _env_render(o,mode='rgb_array',shape_map=None):
    image=o.render(mode=mode).copy()
    if shape_map is None: return image
    return _fix_shape_map(image,shape_map['image'])

def _env_step(o,action,*args,shape_map=None,is_int=False,**kwargs):
    a=action.detach().cpu().numpy()[0]
    if np.isscalar(a):
        a=int(a) if is_int else float(a)
    s,r,d,info=o.step(a,*args,**kwargs)
    s=_fix_shape_map(s,shape_map['state'])
    r=_fix_shape_map(r,shape_map['reward'])
    d=_fix_shape_map(d,shape_map['done'])
    return (s,r,d,info)

def _batchwise_zero(shape,bs): return torch.zeros((bs,*shape[1:]))

class GymSrc(SrcCallback):
    def initialize(self):
        self.source.histories,self.source.pool=L((deque(maxlen=self.steps_count),
                          gym.make(self.env,**self.env_kwargs))
                          for _ in range(self.n_envs)).zip().map(L)
        if self.source.agent is None:
            test_cb=TstCallback(action_space=self.pool[0].action_space)
            self.source.agent=Agent(cbs=test_cb)
        self.action_is_int=False
       # Extra fields
        self.source.next_state=None
        self.source.state=None
        self.source.image=None
        self.source.action=None
        self.source.episode_num=None
        self.source.accum_rewards=None
        self.init_shapes()
        self('reset')
        if self.mode is not None: self.init_render_shapes()
        self.source.imask=torch.zeros((self.n_envs,)).bool()

    def init_render_shapes(self):
        "Set the image shapes with a batch dim."
        image_shape=_env_render(self.pool[0],self.mode).shape
        self.shape_map['image']=_add_batch(image_shape)

    def init_shapes(self):
        "Set the reward shape, state shapes."
        self.shape_map['reward']=(1,1)
        obs_shape=_add_batch(self.pool[0].observation_space.shape)
        for k in ('state','next_state'): self.shape_map[k]=obs_shape
        action=self.pool[0].action_space.sample()
        if isinstance(action,(int,float)):
            self.shape_map['action']=(1,1)
            if isinstance(action,int): self.action_is_int=True
        else:
            self.shape_map['action']=_add_batch(action)

    def reset(self):
        if self.imask.sum()==0:
            _shape_map={k:v for k,v in self.shape_map.items() if k!='episode_num'}
            reset_exps=D(_shape_map).mapv(_batchwise_zero,bs=self.n_envs)
            for k,v in reset_exps.items(): setattr(self.source,k,v)
            self.source.pool.map(_env_seed,seed=self.seed)
            self.source.state=self.pool.map(_env_reset,shape_map=self.shape_map)
            self.source.state=TensorBatch.vstack(tuple(self.state))
            self.source.done=self.source.done.bool()
            self.source.all_done=self.source.all_done.bool()
            self.source.env_id=self.source.env_id.long()
            self.source.p_id=self.source.p_id.long()
            if self.mode is not None:
                self.source.image=self.pool.map(_env_render,mode=self.mode)
            if self.action_is_int:
                self.source.action=self.source.action.long()

    def after_reset(self):
        if self.imask.sum()==0:
            if self.source.episode_num is None:
                self.source.episode_num=self.source.env_id.detach().clone()+1
            else:
                self.source.episode_num+=self.source.episode_num.max()+1-self.source.episode_num
            self.source.all_exp=deepcopy(self.active_exp(self.n_envs))
            self.source.imask=torch.ones((self.n_envs,)).bool()

    def do_action(self):
        self.source.action,exp=self.agent.do_action(**self.active_exp())
        # TODO: Im not sure we need this section, maybe continuous actions need this here though?
#         self.source.action=_fix_shape_map(self.action,self.shape_map['action'],bs=int(self.imask.sum()))
        if not isinstance(self.action,TensorBatch):
            self.source.action=TensorBatch(self.source.action)
        if self.action_is_int:
            self.source.action=self.source.action.long()
        for k in exp:
            if k=='action': continue
            setattr(self.source,k,exp[k])

    def do_step(self):
        step_res=self.pool[self.imask].zipwith(self.action).starmap(_env_step,
                                                                    shape_map=self.shape_map,
                                                                    is_int=self.action_is_int)
        next_states,rewards,dones=step_res.zip()[:3]
        if self.mode is not None:
            self.source.image=self.pool[self.imask].map(_env_render,mode=self.mode)
        self.source.next_state=TensorBatch.vstack(next_states)
        self.source.reward=TensorBatch.vstack(rewards)
        self.source.accum_rewards+=TensorBatch.vstack(rewards)
        self.source.done=TensorBatch.vstack(dones).bool()
        running_mask=self.imask.nonzero().reshape(-1,1)
        self.source.env_id=TensorBatch(running_mask,bs=self.imask.sum()).long()
        worker_id=get_worker_info()
        worker_id=worker_id.id if worker_id is not None else 0
        self.source.p_id=TensorBatch(torch.full(running_mask.shape,worker_id),bs=self.imask.sum()).long()

    def history(self):
        if self.current_history['all_done'].shape[0]==1 and self.current_history['done'][0][0]:
#             print(self.current_history)
            self.current_history['all_done'][0][0]=True

    def after_history(self):
        active_exp=self.active_exp()
        for k in self.all_exp:
            if k in ['episode_num']:continue
            self.all_exp[k][self.imask]=torch.clone(active_exp[k])
        self.source.imask=~self.all_exp['done'].reshape(-1,)
        for k in self.all_exp:
            if k in ['episode_num']:continue
            setattr(self.source,k,self.all_exp[k][self.imask])
        self.source.state=torch.clone(self.next_state)
        self.source.step+=1

# Cell
class SrcProfiler(SrcCallback):
    "Checks time between callbacks."
    def __init__(self):
        self.callback_time:Dict={}
        self.callback_deltas:Dict={}

    def before_set(self,callback_name):
        self.callback_time[callback_name]=datetime.now()
    def after_set_delta(self,callback_name):
        delta=(self.callback_time[callback_name]-datetime.now())
        seconds=0 if delta.seconds==86399 else delta.seconds
        if callback_name not in self.callback_deltas:self.callback_deltas[callback_name]=0
        self.callback_deltas[callback_name]+=seconds+delta.microseconds/1000000

    def before_initialize(self): self.before_set('initialize')
    def after_initialize(self):  self.after_set_delta('initialize')
    def before_reset(self): self.before_set('reset')
    def after_reset(self):  self.after_set_delta('reset')
    def before_do_step(self): self.before_set('do_step')
    def after_do_step(self):  self.after_set_delta('do_step')
    def before_render(self): self.before_set('render')
    def after_render(self):  self.after_set_delta('render')
    def before_history(self): self.before_set('history')
    def after_history(self):  self.after_set_delta('history')
    def before_do_action(self): self.before_set('do_action')
    def after_do_action(self):  self.after_set_delta('do_action')

    def relative_time(self):
        max_v=max(self.callback_deltas.values())
        return D(self.callback_deltas).mapv(lambda o:o/max_v)

# Cell
class FirstLast(SrcCallback):
    gamma=0.99
    def history(self):
        element=self.current_history[0]
        if self.current_history.bs()!=1:
            remainder=self.current_history[1:]
            reward=element['reward']
            for e in reversed(remainder['reward']):
                reward*=self.gamma
                reward+=e
            element['reward']=reward
            element['next_state']=self.current_history[-1]['next_state']
            element['done']=self.current_history[-1]['done']
            self.source.current_history=element
        else:
            self.source.current_history=element

# Cell
class SourceDataset(IterableDataset):
    "Iterates through a `source` object. Allows for re-initing source connections when `num_workers>0`"
    def __init__(self,source=None):
        if not is_listy(source): self.source=source
        elif is_iter(source[0]): self.source=list(map(iter,self.source))
        else:                    self.source=source

    def __iter__(self):          return iter(self.source)
    def wif(self):               self.source._init_state()

# Cell
class IterableTfmdLists(TfmdLists):
    def _after_item(self, o): return self.tfms(next(o))

# Cell
from torch.utils.data._utils.collate import default_collate_err_msg_format
from torch.utils.data._utils.collate import *

def vstack_collate(batch):
    "99% similar to default_collate, however vstacks tensors thus assuming they already have a batch dim"

    elem = batch[0]
    elem_type = type(elem)
    if isinstance(elem, torch.Tensor):
        out = None
        if torch.utils.data.get_worker_info() is not None:
            # If we're in a background process, concatenate directly into a
            # shared memory tensor to avoid an extra copy
            numel = sum([x.numel() for x in batch])
            storage = elem.storage()._new_shared(numel)
            out = elem.new(storage)
        return torch.vstack(batch, out=out)
    elif elem_type.__module__ == 'numpy' and elem_type.__name__ != 'str_' \
            and elem_type.__name__ != 'string_':
        if elem_type.__name__ == 'ndarray' or elem_type.__name__ == 'memmap':
            # array of string classes and object
            if np_str_obj_array_pattern.search(elem.dtype.str) is not None:
                raise TypeError(default_collate_err_msg_format.format(elem.dtype))

            return vstack_collate([torch.as_tensor(b) for b in batch])
        elif elem.shape == ():  # scalars
            return torch.as_tensor(batch)
    elif isinstance(elem, float):
        return torch.tensor(batch, dtype=torch.float64)
    elif isinstance(elem, int_classes):
        return torch.tensor(batch)
    elif isinstance(elem, string_classes):
        return batch
    elif isinstance(elem, container_abcs.Mapping):
        return {key: vstack_collate([d[key] for d in batch]) for key in elem}
    elif isinstance(elem, tuple) and hasattr(elem, '_fields'):  # namedtuple
        return elem_type(*(default_collate(samples) for samples in zip(*batch)))
    elif isinstance(elem, container_abcs.Sequence):
        # check to make sure that the elements in batch have consistent size
        it = iter(batch)
        elem_size = len(next(it))
        if not all(len(elem) == elem_size for elem in it):
            raise RuntimeError('each element in list of batch should be of equal size')
        transposed = zip(*batch)
        return [vstack_collate(samples) for samples in transposed]

    raise TypeError(default_collate_err_msg_format.format(elem_type))

# Cell
_collate_types = (ndarray, Tensor, typing.Mapping, str)

def fr_collate(t):
    "A replacement for PyTorch `default_collate` which maintains types and handles `Sequence`s"
    b = t[0]
    return (vstack_collate(t) if isinstance(b, _collate_types)
            else type(t[0])([fr_collate(s) for s in zip(*t)]) if isinstance(b, Sequence)
            else vstack_collate(t))

# Cell
from fastai.data.all import *

class IterableTfmdDL(TfmdDL):
    def create_batch(self, b): return (fr_collate,fa_convert)[self.prebatched](b)

# Cell
class IterableDataBlock(DataBlock):
    tl_type = TfmdLists

    @delegates(DataBlock)
    def __init__(self,blocks=None, **kwargs):
        blocks = L(self.blocks if blocks is None else blocks)
        for b in blocks:
            if getattr(b, 'tl_type', None) is not None: self.tl_type = b.tl_type
        super().__init__(blocks=blocks, **kwargs)

    def datasets(self, source, verbose=False):
        self.source = source                     ; pv(f"Collecting items from {source}", verbose)
        items = (self.get_items or noop)(source) ; pv(f"Found {len(items)} items", verbose)
        splits = (self.splitter or RandomSplitter())(items)
        pv(f"{len(splits)} datasets of sizes {','.join([str(len(s)) for s in splits])}", verbose)
        return Datasets(items, tfms=self._combine_type_tfms(), splits=splits, dl_type=self.dl_type, n_inp=self.n_inp, verbose=verbose,
                        tl_type=self.tl_type)

# Cell
@patch
def __iter__(self:Datasets):
    for i in cycle(range(len(self))): yield self[i]

@patch
def __init__(self:Datasets, items=None, tfms=None, tls=None, n_inp=None, dl_type=None,tl_type=TfmdLists, **kwargs):
    super(Datasets,self).__init__(dl_type=dl_type)
    self.tls = L(tls if tls else [tl_type(items, t, **kwargs) for t in L(ifnone(tfms,[None]))])
    self.n_inp = ifnone(n_inp, max(1, len(self.tls)-1))

# Cell
class TransformBlock():
    "A basic wrapper that links defaults transforms for the data block API"
    def __init__(self, type_tfms=None, item_tfms=None, batch_tfms=None, dl_type=None, dls_kwargs=None, tl_type=None):
        self.type_tfms  =            L(type_tfms)
        self.item_tfms  = ToTensor + L(item_tfms)
        self.batch_tfms =            L(batch_tfms)
        self.dl_type,self.dls_kwargs = dl_type,({} if dls_kwargs is None else dls_kwargs)
        self.tl_type = tl_type

# Cell
def _load_wif(f): return f

class SourceBlock(TransformBlock):
    def __init__(self,source,dls_kwargs=None,**kwargs):
        dls_kwargs=ifnone(dls_kwargs,{})
        dls_kwargs['indexed']=False
        dls_kwargs['shuffle']=False
        dls_kwargs['wif']=partial(source._init_state)
        dls_kwargs['persistent_workers']=True
        super().__init__(dls_kwargs=dls_kwargs,dl_type=IterableTfmdDL,tl_type=IterableTfmdLists,**kwargs)

def map_iters(_,ls): return list(map(iter,ls))
def noop_splitter(_,o): return [[0]]
def get_all_keys(_,o): return o

class SourceDataBlock(IterableDataBlock):
    get_items=map_iters
    splitter=noop_splitter
    get_y=get_all_keys

    @delegates(DataBlock)
    def __init__(self,blocks=None,n_inp=None,**kwargs):
        if not is_listy(blocks) or len(blocks)==1: n_inp=0
        super().__init__(blocks=blocks,n_inp=n_inp,**kwargs)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/10c_agents.dqn.double.ipynb (unless otherwise specified).

__all__ = ['DoubleDQNTrainer']

# Cell
# Python native modules
import os
from collections import deque
# Third party libs
import torch
from torch.nn import *
from fastcore.all import *
from fastai.learner import *
from fastai.torch_basics import *
from fastai.torch_core import *
from fastai.callback.all import *
# Local modules
from ...data.block_simple import *
from ...data.gym import *
from ...agent import *
from ...core import *
from .core import *
from .targets import *

# Cell
class DoubleDQNTrainer(DQNTargetTrainer):
    def after_pred(self):
        self.learn.yb=self.xb
        self.learn.xb=self.xb[0]

        self._xb=({k:v.clone() for k,v in self.xb.items()},)
        self.learn.done_mask=self.xb['done'].reshape(-1,)
        chosen_actions=self.learn.next_q=self.model.model(self.xb['next_state']).argmax(dim=1).reshape(-1,1)
        self.learn.next_q=self.target_model(self.xb['next_state']).gather(1,chosen_actions)
        self.learn.next_q[self.done_mask]=0
        self.learn.targets=self.xb['reward']+self.learn.next_q*(self.discount**self.n_steps)
        self.learn.pred=self.learn.model.model(self.xb['state'])
        t_q=self.pred.clone()
        t_q.scatter_(1,self.xb['action'],self.targets)
        self.learn.yb=(t_q,)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/11a_agents.policy_gradient.ppo.ipynb (unless otherwise specified).

__all__ = ['Critic', 'ClampBlock', 'show_multisurface', 'ConstBlock', 'clamp_or_const', 'show_surface', 'SimpleGMM',
           'MultiCompGMM']

# Cell
# Python native modules
import os
# Third party libs
import torch
from torch.nn import *
from torch.distributions import *
from fastcore.all import *
from fastai.learner import *
from fastai.torch_basics import *
from fastai.torch_core import *
from fastai.callback.all import *
# Local modules

# Cell
class Critic(Module):
    def __init__(self,state_sz:int,action_sz:int=0,hidden=512):
        "Init model that estimates a single q value based on `state_sz` optionally + `action_sz`"
        self.v=nn.Sequential(
            nn.Linear(state_sz+action_sz,hidden),
            nn.ReLU(),
            nn.Linear(hidden,hidden),
            nn.ReLU(),
            nn.Linear(hidden,1),
        )

    def forward(self, x): return self.v(x.float())

# Cell
class ClampBlock(Module):
    def __init__(self,in_sz:int,out_sz:int,activation_cls=nn.Softplus,clamp_min=0.3,clamp_max=10.0):
        "Linear Layer with clamped outputs."
        store_attr()
        self.fc=nn.Linear(self.in_sz,self.out_sz)
        self.activation=self.activation_cls()

    def forward(self,x): return torch.clamp(self.activation(self.fc(x)),
                                            self.clamp_min,self.clamp_max)

# Cell
def show_multisurface(surfaces,title=''):
    from IPython.display import HTML
    import plotly.graph_objects as go
    from plotly.subplots import make_subplots

    rows=(len(surfaces)+1)//2
    cols=(len(surfaces)+1)%2+1
    fig=make_subplots(rows=rows,cols=cols,
                      specs=[[{"type": "surface"} for _ in range(cols)]]*rows,
                      column_widths=[0.5,0.5])

    for i in range(0,len(surfaces)):
        fig.add_trace(go.Surface(z=to_detach(surfaces[i]),showscale=False),
                      row=(i+2)//2,col=i%2+1)
    fig.update_layout(title=title,autosize=False,
                      width=1000, height=500*rows,#template='plotly_dark',
                      margin=dict(l=50, r=50, b=80, t=90))
    return HTML(fig.to_html())

# Cell
class ConstBlock(Module):
    def __init__(self,out_sz:int,constant=1.0,**kwargs):
        "Module that produces a constant output."
        store_attr()

    def forward(self,x): return torch.full((x.shape[0],self.out_sz),self.constant)

# Cell
@delegates(ClampBlock)
def clamp_or_const(fix_variance:bool,**kwargs):
    return ConstBlock(**kwargs) if fix_variance else ClampBlock(**kwargs)

# Cell
def show_surface(surface,title=''):
    from IPython.display import HTML
    import plotly.graph_objects as go
    from plotly.subplots import make_subplots

    fig=go.Figure()
    fig.add_trace(go.Surface(z=to_detach(surface),showscale=False))
    fig.update_layout(title=title,autosize=False,
                      width=500, height=500,#template='plotly_dark',
                      margin=dict(l=50, r=50, b=80, t=90))
    return HTML(fig.to_html())

# Cell
class SimpleGMM(Module):
    def __init__(self,in_sz,out_sz,fix_variance:bool=False):
        "Single-component GMM parameterized by a fully connected layer with optional std layer."
        store_attr()
        self.mean_fc=nn.Linear(self.in_sz,self.out_sz)
        self.std_fc=clamp_or_const(fix_variance,in_sz=self.in_sz,out_sz=self.out_sz)

    def forward(self,x)->Independent:
        return Independent(Normal(self.mean_fc(x),self.std_fc(x)),1)

# Cell
class MultiCompGMM(Module):
    @delegates(ClampBlock)
    def __init__(self,in_sz,out_sz,n_components,fix_variance:bool=False,**kwargs):
        "Multi-component GMM parameterized by a fully connected layer with optional std layer."
        store_attr()
        self.logit_fc=nn.Linear(self.in_sz,self.n_components)
        self.mean_fcs=nn.ModuleList([nn.Linear(self.in_sz,self.out_sz)
                                     for _ in range(self.n_components)])
        self.std_fcs=nn.ModuleList([clamp_or_const(fix_variance,in_sz=self.in_sz,out_sz=self.out_sz)
                                    for _ in range(self.n_components)])
        self.means,self.logits,self.stds=[],[],[]
        self.log_softmax=LogSoftmax(dim=1)

    def forward(self,x)->MixtureSameFamily:
        self.means=torch.stack([o(x) for o in self.mean_fcs],dim=1)
        self.stds=torch.stack([o(x) for o in self.std_fcs],dim=1)
        self.logits=self.logit_fc(x)
        self.logits=self.log_softmax(self.logits)
        return MixtureSameFamily(
            mixture_distribution=Categorical(logits=self.logits),
            component_distribution=Independent(Normal(self.means,self.stds),1)
        )


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/10a_agents.dqn.core.ipynb (unless otherwise specified).

__all__ = ['DQN', 'ArgMaxFeed', 'DiscreteEpsilonRandomSelect', 'Epsilon', 'ExperienceReplay', 'DQNTrainer']

# Cell
# Python native modules
import os
from collections import deque
# Third party libs
import torch
from torch.nn import *
from fastcore.all import *
from fastai.learner import *
from fastai.torch_basics import *
from fastai.torch_core import *
from fastai.callback.all import *
# Local modules
from ...data.block_simple import *
from ...data.gym import *
from ...agent import *
from ...core import *

# Cell
class DQN(Module):
    def __init__(self,state_sz:int,action_sz:int,hidden=512):
        self.layers=Sequential(
            Linear(state_sz,hidden),
            ReLU(),
            Linear(hidden,action_sz),
        )
    def forward(self,x): return self.layers(x)

# Cell
class ArgMaxFeed(AgentCallback):
    def before_action(self):
        raw_action=self.agent.model(self.experience['state'].to(default_device()))
        self.agent.raw_action_shape=raw_action.shape
        self.agent.action=torch.argmax(raw_action,dim=1).reshape(-1,1)

class DiscreteEpsilonRandomSelect(AgentCallback):

    def __init__(self,epsilon=0.5,idx=0,min_epsilon=0.2,max_epsilon=1,max_steps=5000):
        store_attr()

    def before_noise(self):
        self.mask=torch.randn(size=(self.agent.action.shape[0],))<self.epsilon
        self.experience['randomly_selected']=self.mask.reshape(-1,1)
        self.experience['epsilon']=torch.full(self.agent.action.shape,self.epsilon)
        self.experience['orignal_actions']=self.agent.action.detach().clone()
        self.agent.action[self.mask]=self.agent.action[self.mask].random_(0,self.agent.raw_action_shape[1])
        self.agent.action=self.agent.action.detach().cpu().numpy()

        if self.agent.model.training:
            self.idx+=1
            self.epsilon=max(self.min_epsilon,self.max_epsilon-self.idx/self.max_steps)

# Cell
class Epsilon(Metric):
    order=30
    epsilon=0

    @property
    def value(self): return self.epsilon
    def reset(self): self.epsilon=0
    def accumulate(self,learn):
        for cb in learn.model.cbs:
            if type(cb)==DiscreteEpsilonRandomSelect:
                self.epsilon=cb.epsilon

# Cell
class ExperienceReplay(Callback):
    def __init__(self,*args,bs=16,max_sz=200,warmup_sz=100,**kwargs):
        store_attr()
        self.memory=None
        self.pointer=0

    def after_pred(self):
        xb=BD(self.learn.xb[0]).mapv(to_detach)
        if self.memory is None:            self.memory=xb
        elif self.memory.bs()<self.max_sz: self.memory+=xb
        else:
            self.memory=self.memory[:self.pointer]+xb+self.memory[self.pointer+xb.bs():]
            self.pointer+=xb.bs()
            if self.pointer>self.max_sz: self.pointer=0
        with torch.no_grad():
            idxs=np.random.randint(0,self.memory.bs(),self.bs).tolist()
            self.learn.xb=(self.memory[idxs].mapv(to_device),)

        if self.memory.bs()<self.warmup_sz: raise CancelBatchException

# Cell
class DQNTrainer(Callback):
    "Performs traditional training on `next_q`. Requires a callback such as `RegularNextQ`"
    def __init__(self,discount=0.99,n_steps=1):
        store_attr()
        self._xb=None

    def after_pred(self):
        self.learn.yb=self.xb
        self.learn.xb=self.xb[0]
        self._xb=({k:v.clone() for k,v in self.xb.items()},)
        self.learn.done_mask=self.xb['done'].reshape(-1,)
        self.learn.next_q=self.learn.model.model(self.xb['next_state']).max(dim=1).values.reshape(-1,1)
        self.learn.next_q[self.done_mask]=0 #xb[done_mask]['reward']
        self.learn.targets=self.xb['reward']+self.learn.next_q*(self.discount**self.n_steps)
        self.learn.pred=self.learn.model.model(self.xb['state'])
        t_q=self.pred.clone()
        t_q.scatter_(1,self.xb['action'],self.targets)
        # finalize the xb and yb
        self.learn.yb=(t_q,)

    def before_backward(self):
        self.learn.xb=self._xb


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/20_test_utils.ipynb (unless otherwise specified).

__all__ = ['show_install']

# Cell
# Python native modules
import os
# Third party libs
from fastcore.all import *
from fastai.test_utils import *
import torch
# Local modules

# Cell

#export
def show_install(show_nvidia_smi:bool=False):
    "Print user's setup information"

    import fastai, platform, fastprogress, fastcore, fastrl

    rep = []
    opt_mods = []

    rep.append(["=== Software ===", None])
    rep.append(["python", platform.python_version()])
    rep.append(["fastrl", fastrl.__version__])
    rep.append(["fastai", fastai.__version__])
    rep.append(["fastcore", fastcore.__version__])
    rep.append(["fastprogress", fastprogress.__version__])
    rep.append(["torch",  torch.__version__])

    # nvidia-smi
    smi = nvidia_smi()
    if smi:
        match = re.findall(r'Driver Version: +(\d+\.\d+)', smi)
        if match: rep.append(["nvidia driver", match[0]])

    available = "available" if torch.cuda.is_available() else "**Not available** "
    rep.append(["torch cuda", f"{torch.version.cuda} / is {available}"])

    # no point reporting on cudnn if cuda is not available, as it
    # seems to be enabled at times even on cpu-only setups
    if torch.cuda.is_available():
        enabled = "enabled" if torch.backends.cudnn.enabled else "**Not enabled** "
        rep.append(["torch cudnn", f"{torch.backends.cudnn.version()} / is {enabled}"])

    rep.append(["\n=== Hardware ===", None])

    gpu_total_mem = []
    nvidia_gpu_cnt = 0
    if smi:
        mem = nvidia_mem()
        nvidia_gpu_cnt = len(ifnone(mem, []))

    if nvidia_gpu_cnt: rep.append(["nvidia gpus", nvidia_gpu_cnt])

    torch_gpu_cnt = torch.cuda.device_count()
    if torch_gpu_cnt:
        rep.append(["torch devices", torch_gpu_cnt])
        # information for each gpu
        for i in range(torch_gpu_cnt):
            rep.append([f"  - gpu{i}", (f"{gpu_total_mem[i]}MB | " if gpu_total_mem else "") + torch.cuda.get_device_name(i)])
    else:
        if nvidia_gpu_cnt:
            rep.append([f"Have {nvidia_gpu_cnt} GPU(s), but torch can't use them (check nvidia driver)", None])
        else:
            rep.append([f"No GPUs available", None])


    rep.append(["\n=== Environment ===", None])

    rep.append(["platform", platform.platform()])

    if platform.system() == 'Linux':
        distro = try_import('distro')
        if distro:
            # full distro info
            rep.append(["distro", ' '.join(distro.linux_distribution())])
        else:
            opt_mods.append('distro');
            # partial distro info
            rep.append(["distro", platform.uname().version])

    rep.append(["conda env", get_env('CONDA_DEFAULT_ENV')])
    rep.append(["python", sys.executable])
    rep.append(["sys.path", "\n".join(sys.path)])

    print("\n\n```text")

    keylen = max([len(e[0]) for e in rep if e[1] is not None])
    for e in rep:
        print(f"{e[0]:{keylen}}", (f": {e[1]}" if e[1] is not None else ""))

    if smi:
        if show_nvidia_smi: print(f"\n{smi}")
    else:
        if torch_gpu_cnt: print("no nvidia-smi is found")
        else: print("no supported gpus found on this system")

    print("```\n")

    print("Please make sure to include opening/closing ``` when you paste into forums/github to make the reports appear formatted as code sections.\n")

    if opt_mods:
        print("Optional package(s) to enhance the diagnostics can be installed with:")
        print(f"pip install {' '.join(opt_mods)}")
        print("Once installed, re-run this utility to get the additional information")


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/05_data.block.ipynb (unless otherwise specified).

__all__ = ['DQN', 'TestDataset']

# Cell
# Python native modules
import os
# Third party libs
from fastcore.all import *
from fastai.torch_basics import *
from fastai.data.all import *
from fastai.basics import *
from torch.utils.data import Dataset
from torch import nn
# Local modules

# Cell
class DQN(Module):
    def __init__(self):
        self.policy=nn.Sequential(
            nn.Linear(4,50),
            nn.ReLU(),
            nn.Linear(50,2)
        )

    def forward(self,x):
        return torch.argmax(self.policy(x),dim=0)

# Cell
class TestDataset(Dataset):
    def __init__(self,policy,device='cpu'):
        self.policy=policy
        self.env=gym.make('CartPole-v1')
        self.next_state=self.env.reset()
        self.device='cpu'

    def __len__(self): return 100
    def __getitem__(self,idx):
        print(id(self.env),' ')
        self.next_state, r, is_done, _=self.env.step(self.policy(Tensor(self.next_state).to(self.device)).numpy())
        if is_done:self.next_state=self.env.reset()
        return self.next_state

# Cell
# Python native modules
import os
# Third party libs
from fastcore.all import *
from fastai.torch_basics import *
from fastai.data.all import *
from fastai.basics import *
from torch.utils.data import Dataset
from torch import nn
# Local modules
from .block import *


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/03_callback.core.ipynb (unless otherwise specified).

__all__ = ['Loop', 'loop_event', 'LoopCallback']

# Cell
# Python native modules
import os
from typing import *
# Third party libs
from fastcore.all import *
from fastai.callback.core import *
from fastai.data.all import *
# Local modules

# Cell
defaults.loop_callbacks=L()

_loop = ['Start Fit', 'before_fit', 'Start Epoch Loop', 'before_epoch', 'Start Train', 'before_train',
         'Start Batch Loop', 'before_batch', 'after_pred', 'after_loss', 'before_backward', 'before_step',
         'after_step', 'after_cancel_batch', 'after_batch','End Batch Loop','End Train',
         'after_cancel_train', 'after_train', 'Start Valid', 'before_validate','Start Batch Loop',
         '**CBs same as train batch**', 'End Batch Loop', 'End Valid', 'after_cancel_validate',
         'after_validate', 'End Epoch Loop', 'after_cancel_epoch', 'after_epoch', 'End Fit',
         'after_cancel_fit', 'after_fit']

class Loop(GetAttr):
    _loop,_default,_events=_loop,'_base',None
    def __init__(self,cbs=None,**kwargs):
        store_attr(but='cbs')
        self.cbs=L()
        self.add_cbs(L(defaults.loop_callbacks)+L(cbs))
        self("after_create")

    def _grab_cbs(self, cb_cls): return L(cb for cb in self.cbs if isinstance(cb, cb_cls))

    def add_cbs(self, cbs):
        L(cbs).map(self.add_cb)
        return self

    def remove_cbs(self, cbs):
        L(cbs).map(self.remove_cb)
        return self

    def add_cb(self, cb):
        if isinstance(cb, type): cb = cb()
        setattr(cb,self._default,self)
        setattr(self, cb.name, cb)
        self.cbs.append(cb)
        return self

    def remove_cb(self, cb):
        if isinstance(cb, type): self.remove_cbs(self._grab_cbs(cb))
        else:
            setattr(cb,self._default,None)
            if hasattr(self, cb.name): delattr(self, cb.name)
            if cb in self.cbs: self.cbs.remove(cb)
        return self

    def _with_events(self, f, event_type, ex, final=noop):
        try: self(f'before_{event_type}');  f()
        except ex: self(f'after_cancel_{event_type}')
        self(f'after_{event_type}');  final()

    @contextmanager
    def added_cbs(self, cbs):
        self.add_cbs(cbs)
        try: yield
        finally: self.remove_cbs(cbs)

    @contextmanager
    def removed_cbs(self, cbs):
        self.remove_cbs(cbs)
        try: yield self
        finally: self.add_cbs(cbs)

    def ordered_cbs(self, event): return [cb for cb in self.cbs.sorted('order') if hasattr(cb, event)]
    def __call__(self, event_name): L(event_name).map(self._call_one)

    def _call_one(self, event_name):
        if not hasattr(self._events, event_name): raise Exception(f'missing {event_name}')
        for cb in self.cbs.sorted('order'): cb(event_name)


    def show_loop(self):
        indent = 0
        for s in self._loop:
            s=s.replace('event.','')
            if s.startswith('Start'): print(f'{" "*indent}{s}'); indent += 2
            elif s.startswith('End'): indent -= 2; print(f'{" "*indent}{s}')
            else: print(f'{" "*indent} - {s:15}:', self.ordered_cbs(s))

# Cell
_events = L.split('')

mk_class('loop_event', **_events.map_dict(),
         doc="All possible events as attributes to get tab-completion and typo-proofing")

#nbdev_comment _all_ = ['loop_event']

# Cell
_inner_loop = "".split()

# Cell
@funcs_kwargs(as_method=True)
class LoopCallback(Stateful,GetAttr):
    "Basic class handling tweaks of a callback loop by changing a `obj` in various events"
    order,_default,obj,run,run_train,run_valid = 0,'obj',None,True,True,True
    end_event='after_fit'
    _methods = _events

    def __init__(self, **kwargs): assert not kwargs, f'Passed unknown events: {kwargs}'
    def __repr__(self): return type(self).__name__

    def __call__(self, event_name,**kwargs):
        "Call `self.{event_name}` if it's defined"
        _run = (event_name not in _inner_loop or (self.run_train and getattr(self, 'training', True)) or
               (self.run_valid and not getattr(self, 'training', False)))
        res=None
        if self.run and _run:
            res=getattr(self, event_name, noop)(**kwargs)
        if event_name==self.end_event: self.run=True #Reset self.run to True at each end of fit
        return res

    def __setattr__(self, name, value):
        if hasattr(self.obj,name):
            warn(f"You are shadowing an attribute ({name}) that exists in the {self._default}. Use `self.{self._default}.{name}` to avoid this")
        super().__setattr__(name, value)

    @property
    def name(self):
        "Name of the `LoopCallback`, camel-cased and with '*LoopCallback*' removed"
        return class2attr(self, 'LoopCallback')


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/04_agent.ipynb (unless otherwise specified).

__all__ = ['agent_event', 'Agent', 'AgentCallback']

# Cell
# Python native modules
import os
# Third party libs
from fastcore.all import *
# Local modules
from .callback.core import *

# Cell
_events = L.split('create preprocess action noise')
_events = _events.map(lambda s:'after_cancel_'+s)+ _events.map(lambda s:'after_cancel_'+s) + _events.map(lambda s:'before_'+s) + _events.map(lambda s:'after_'+s)

mk_class('agent_event', **_events.map_dict(),
         doc="All possible events as attributes to get tab-completion and typo-proofing")

#nbdev_comment _all_ = ['agent_event']

_inner_loop = "".split()

# Cell
class Agent(Loop):
    _loop=L(['Start Action','before_preprocess','after_preprocess',
                            'before_action','after_action',
                            'before_noise','after_noise','End Action'])
    _default='agent'
    end_event='after_noise'
    @delegates(Loop)
    def __init__(self,model=None,**kwargs):
        super().__init__(**kwargs)
        self.experience=None
        self.action=None
        self.model=model

    def state_dict(self): return self.model.state_dict()
    def parameters(self): return self.model.parameters()
    def children(self): return self.model.children()
    def train(self,*args,**kwargs): return self.model.train(*args,**kwargs)
    def to(self,*args,**kwargs):
        self.model.to(*args,**kwargs)
        return self

    def _call_one(self, event_name):
        if not hasattr(agent_event, event_name): raise Exception(f'missing {event_name}')
        for cb in self.cbs.sorted('order'): cb(event_name)

    def __call__(self, event_name=None,**kwargs):
        if isinstance(event_name,str): L(event_name).map(self._call_one)
        elif isinstance(event_name,dict):return self.do_action(**event_name)

    def do_action(self,**kwargs):
        self.experience=kwargs
        for event in L.split('preprocess action noise'):
            self._with_events(noop,event,Exception)
        return self.action,self.experience

# Cell
class AgentCallback(LoopCallback):
    _default,agent='agent',None
    _methods=_events

    def after_cancel_preprocess(self):raise
    def after_cancel_action(self):    raise
    def after_cancel_noise(self):     raise


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/00_nbdev_extension.ipynb (unless otherwise specified).

__all__ = ['dependencies', 'create_conda_yaml', 'create_conda_yamls']

# Cell
def dependencies(dev:bool=False,cfg_name='settings.ini'):
    "Gets a list of dependencies in a `cfg_name` for conda compatability."
    c=Config(cfg_name)
    deps=[f'python={c.min_python}','pip','setuptools']
    if c.requirements:             deps+=c.requirements.split(' ')
    if c.conda_requirements:       deps+=c.conda_requirements.split(' ')
    if dev and c.dev_requirements: deps+=c.dev_requirements.split(' ')
    if c.pip_requirements:         deps+=[{'pip':c.pip_requirements.split(' ')}]
    return deps

# Cell
def create_conda_yaml(channels:str='conda-forge,pytorch,fastai',
                      cfg_name='settings.ini',dev:bool=False):
    "Creates a conda dictionary of the format of an env file."
    c=Config()
    return {'name':c.lib_name if not dev else c.lib_name+'_dev',
            'channels': channels.split(','),
            'dependencies': dependencies(dev=dev,cfg_name=cfg_name)}

# Cell
def create_conda_yamls(also_dev:bool=True,cfg_name='settings.ini',sub_dir=''):
    "Creates conda env for normal and development environments."
    c=Config(cfg_name)
    parent=c.config_path/Path(sub_dir) if sub_dir else c.config_path
    parent.mkdir(parents=True,exist_ok=True)
    for is_dev in ([False,True] if also_dev else [False]):
        fname=(c.lib_name+f'{"_dev" if is_dev else ""}_env.yaml')
        with open(parent/fname,'w') as f:
            d=create_conda_yaml(cfg_name=cfg_name,dev=is_dev)
            yaml.dump(d,f)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/00_core.ipynb (unless otherwise specified).

__all__ = ['map_dict_ex', 'D', 'tensor2shape', 'tensor2mu', 'TensorBatch', 'obj2tensor', 'BD']

# Cell
# Python native modules
import os,warnings
# Third party libs
from fastcore.all import *
from fastai.torch_core import *
from fastai.basics import *
import pandas as pd
import torch
import numpy as np
# Local modules

# Cell
def map_dict_ex(d,f,*args,gen=False,wise=None,**kwargs):
    "Like `map`, but for dicts and uses `bind`, and supports `str` and indexing"
    g = (bind(f,*args,**kwargs) if callable(f)
         else f.format if isinstance(f,str)
         else f.__getitem__)

    if wise is None:  return map(g,d.items())
    return ((k,g(v)) if wise=='value' else (g(k),v) for k,v in d.items())

# Cell
_error_msg='Found idxs: %s have values more than %s e.g.: %s'

class D(dict):
    "Improved version of `dict` with array handling abilities"
    def __init__(self,*args,mapping=False,**kwargs):
        self.mapping=mapping
        super().__init__(*args,**kwargs)

    def eq_k(self,o:'D',with_diff=False):
        eq=set(o.keys())==set(self.keys())
        if with_diff: return eq,set(o.keys()).symmetric_difference(set(self.keys()))
        return eq
    def _new(self,*args,**kwargs): return type(self)(*args,**kwargs)

    def map(self,f,*args,gen=False,**kwargs):
        return (self._new,noop)[gen](map_dict_ex(self,f,*args,**kwargs),mapping=True)
    def mapk(self,f,*args,gen=False,wise='key',**kwargs):
        return self.map(f,*args,gen=gen,wise=wise,**kwargs)
    def mapv(self,f,*args,gen=False,wise='value',**kwargs):
        return self.map(f,*args,gen=gen,wise=wise,**kwargs)

# Cell
def tensor2shape(k,t:'TensorBatch',relative_shape=False):
    "Converts a tensor into a dict of shapes, or a 1d numpy array"
    return {
        k:t.cpu().numpy().reshape(-1,) if len(t.shape)==2 and t.shape[1]==1 else
        [str((1,*t.shape[1:]) if relative_shape else t.shape)]*t.shape[0]
    }

# Cell
def tensor2mu(k,t:Tensor): return {f'{k}_mu':t.reshape(t.shape[0],-1).double().mean(axis=1)}
tensor2mu.__docs__="Returns a dict with key `k`_mu with the mean of `t` batchwise "

# Cell
class TensorBatch(TensorBase):
    "A tensor assumes a batch dimension"
    def __new__(cls, x, bs=1,**kwargs):
        res=super(TensorBatch,cls).__new__(cls,x,**kwargs)
        setattr(res,'_bs',x.bs() if isinstance(x,cls) else bs)
        return res

    def strict(self):
        assert self.shape[0]==self._bs,f'Tensor has shape {self.shape} while bs is {self._bs}'

    def bs(self): return self.shape[0]
    def get(self,*args):
        "Get a possible subset of a tensor while maintaining a batch dim."
        res=self[args]
        if len(self.shape)>len(res.shape): res=res.unsqueeze(0)
        return res

    @classmethod
    def vstack(cls,*args):
        return cls(torch.vstack(*args),bs=L(*args).map(cls).map(Self.bs()).sum())

def obj2tensor(o):
    return (o if isinstance(o,TensorBatch) else
            TensorBatch(o) if isinstance(o,(L,list,np.ndarray,Tensor,TensorBatch)) else
            TensorBatch([o]))

def _get_bs(o): return o.bs if isinstance(o,TensorBatch) else TensorBatch(o).bs

# export
class BD(D):
    def __init__(self,*args,**kwargs):
        super().__init__(*args,**kwargs)
        self.bs=list(self.values())[0].bs

    def __radd__(self,o): return self if isinstance(o,int) else self.__add__(o)
    def __add__(self,o):
        return BD({k:TensorBatch.vstack((self[k],o[k])) for k in self})

    def __getitem__(self,o):
        if is_listy(o) or isinstance(o,(TensorBatch,int,Tensor)):
            return type(self)({k:self[k].get(o) for k in self})
        return super().__getitem__(o)

    @classmethod
    def merge(cls,*ds,**kwargs): return cls(merge(*ds),**kwargs)
    @delegates(pd.DataFrame)
    def pandas(self,mu=False,relative_shape=False,jupyter_nrows=None,**kwargs):
        "Turns a `BD` into a pandas Dataframe optionally showing `mu` of values."
        if jupyter_nrows is not None: pd.set_option('display.max_rows', jupyter_nrows)
        return pd.DataFrame(merge(
            *tuple(tensor2shape(k,v,relative_shape) for k,v in self.items()),
            *(tuple(tensor2mu(k,v) for k,v in self.items()) if mu else ())
        ),**kwargs)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/10d_agents.dqn.dueling.ipynb (unless otherwise specified).

__all__ = ['DuelingBlock', 'DuelingDQN']

# Cell
# Python native modules
# Third party libs
from torch.nn import *
from fastcore.all import *
from fastai.learner import *
from fastai.torch_basics import *
from fastai.torch_core import *
from fastai.callback.all import *
# Local modules
from ...data.block_simple import *
from ...data.gym import *
from ...agent import *
from ...core import *
from .core import *
from .targets import *
from .double import *

# Cell
class DuelingBlock(nn.Module):
    def __init__(self,n_actions,hidden=512,lin_cls=nn.Linear):
        super().__init__()
        self.val=lin_cls(hidden,1)
        self.adv=lin_cls(hidden,n_actions)

    def forward(self,xi):
        val,adv=self.val(xi),self.adv(xi)
        xi=val.expand_as(adv)+(adv-adv.mean()).squeeze(0)
        return xi

class DuelingDQN(DQN):
    def __init__(self,state_sz:int,n_actions,hidden=512):
        super(DQN,self).__init__()
        self.layers=nn.Sequential(
            nn.Linear(state_sz,hidden),
            nn.ReLU(),
            DuelingBlock(n_actions,hidden)
        )


# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"map_dict_ex": "00_core.ipynb",
         "D": "00_core.ipynb",
         "tensor2shape": "00_core.ipynb",
         "tensor2mu": "00_core.ipynb",
         "tensor2mu.__docs__": "00_core.ipynb",
         "TensorBatch": "00_core.ipynb",
         "obj2tensor": "00_core.ipynb",
         "BD": "00_core.ipynb",
         "dependencies": "00_nbdev_extension.ipynb",
         "create_conda_yaml": "00_nbdev_extension.ipynb",
         "create_conda_yamls": "00_nbdev_extension.ipynb",
         "Loop": "03_callback.core.ipynb",
         "defaults.loop_callbacks": "03_callback.core.ipynb",
         "LoopCallback": "03_callback.core.ipynb",
         "Agent": "04_agent.ipynb",
         "AgentCallback": "04_agent.ipynb",
         "TstCallback": "05c_data.gym.ipynb",
         "Src": "05a_data.block.ipynb",
         "SrcCallback": "05a_data.block.ipynb",
         "GymSrc": "05a_data.block.ipynb",
         "SrcProfiler": "05a_data.block.ipynb",
         "FirstLast": "05c_data.gym.ipynb",
         "SourceDataset": "05b_data.block_simple.ipynb",
         "IterableTfmdLists": "05b_data.block_simple.ipynb",
         "vstack_collate": "05b_data.block_simple.ipynb",
         "fr_collate": "05b_data.block_simple.ipynb",
         "IterableTfmdDL": "05b_data.block_simple.ipynb",
         "IterableDataBlock": "05b_data.block_simple.ipynb",
         "Datasets.__iter__": "05a_data.block.ipynb",
         "Datasets.__init__": "05b_data.block_simple.ipynb",
         "TransformBlock": "05b_data.block_simple.ipynb",
         "SourceBlock": "05a_data.block.ipynb",
         "map_iters": "05a_data.block.ipynb",
         "noop_splitter": "05a_data.block.ipynb",
         "get_all_keys": "05a_data.block.ipynb",
         "SourceDataBlock": "05b_data.block_simple.ipynb",
         "parse_events": "05b_data.block_simple.ipynb",
         "Source": "05c_data.gym.ipynb",
         "VerboseChecked": "05b_data.block_simple.ipynb",
         "IterableDatasets": "05b_data.block_simple.ipynb",
         "get_sources": "05b_data.block_simple.ipynb",
         "Callback.after_create": "05b_data.block_simple.ipynb",
         "Recorder.after_batch": "05b_data.block_simple.ipynb",
         "return_data": "05c_data.gym.ipynb",
         "ReturnHistory": "05c_data.gym.ipynb",
         "np_type_cast": "05c_data.gym.ipynb",
         "GymLoop": "05c_data.gym.ipynb",
         "FirstLastTfm": "05c_data.gym.ipynb",
         "Reward": "05c_data.gym.ipynb",
         "NEpisodes": "05c_data.gym.ipynb",
         "DQN": "10a_agents.dqn.core.ipynb",
         "ArgMaxFeed": "10a_agents.dqn.core.ipynb",
         "DiscreteEpsilonRandomSelect": "10a_agents.dqn.core.ipynb",
         "Epsilon": "10a_agents.dqn.core.ipynb",
         "ExperienceReplay": "10a_agents.dqn.core.ipynb",
         "DQNTrainer": "10a_agents.dqn.core.ipynb",
         "DQNTargetTrainer": "10b_agents.dqn.targets.ipynb",
         "DoubleDQNTrainer": "10c_agents.dqn.double.ipynb",
         "DuelingBlock": "10d_agents.dqn.dueling.ipynb",
         "DuelingDQN": "10d_agents.dqn.dueling.ipynb",
         "create_support": "10e_agents.dqn.categorical.ipynb",
         "CategoricalDQN": "10e_agents.dqn.categorical.ipynb",
         "distribute": "10e_agents.dqn.categorical.ipynb",
         "final_distribute": "10e_agents.dqn.categorical.ipynb",
         "categorical_update": "10e_agents.dqn.categorical.ipynb",
         "show_q_distribution": "10e_agents.dqn.categorical.ipynb",
         "PartialCrossEntropy": "10e_agents.dqn.categorical.ipynb",
         "CategoricalDQNTrainer": "10e_agents.dqn.categorical.ipynb",
         "CategoricalArgMaxFeed": "10e_agents.dqn.categorical.ipynb",
         "show_q": "10e_agents.dqn.categorical.ipynb",
         "show_q_and_max_distribution": "10e_agents.dqn.categorical.ipynb",
         "Critic": "11a_agents.policy_gradient.ppo.ipynb",
         "ClampBlock": "11a_agents.policy_gradient.ppo.ipynb",
         "show_multisurface": "11a_agents.policy_gradient.ppo.ipynb",
         "ConstBlock": "11a_agents.policy_gradient.ppo.ipynb",
         "clamp_or_const": "11a_agents.policy_gradient.ppo.ipynb",
         "show_surface": "11a_agents.policy_gradient.ppo.ipynb",
         "SimpleGMM": "11a_agents.policy_gradient.ppo.ipynb",
         "MultiCompGMM": "11a_agents.policy_gradient.ppo.ipynb",
         "show_install": "20_test_utils.ipynb"}

modules = ["core.py",
           "nbdev_extensions.py",
           "callback/core.py",
           "agent.py",
           "data/block.py",
           "data/block_simple.py",
           "data/gym.py",
           "agents/dqn/core.py",
           "agents/dqn/targets.py",
           "agents/dqn/double.py",
           "agents/dqn/dueling.py",
           "agents/dqn/categorical.py",
           "agents/policy_gradient/ppo.py",
           "test_utils.py"]

doc_url = "https://josiahls.github.io/fastrl/"

git_url = "https://github.com/josiahls/fastrl/tree/master/"

def custom_doc_links(name): return None
